"""Module for parsing and validating config files."""
import sys
import importlib

from json.decoder import JSONDecoder, JSONDecodeError

from collectington.logger import setup_logging

DECODER = JSONDecoder()
LOGGER = setup_logging()


def get_config(path):
    """Read config from a file, parse it, and validate it."""
    with open(path, encoding="utf-8") as file:
        contents = file.read()

    config = parse(contents)

    validate(config)  # ValueError will be raised if there is an issue with the config

    return config


def parse(config):
    """Parse the string of a config file."""
    try:
        output = DECODER.decode(config)
    except JSONDecodeError as err:
        LOGGER.error("Failed to parse config file: %s", err)
        raise
    except TypeError as err:
        LOGGER.error("Failed to parse config file: %s", err)
        raise

    return output


def get_list_of_available_metrics(config, service_name):
    """
    Get all the available metrics from a config file
    and merge it into a single list.
    """
    metrics_mapping = config["services"][service_name][
        "prometheus_metrics_mapping"
    ].values()
    list_of_available_metrics = [
        item for sublist in metrics_mapping for item in sublist
    ]

    return list_of_available_metrics


def get_service(config, service_name):
    """Get service class instance using config"""
    service = config["services"][service_name]["service_class"]
    service_module = config["services"][service_name]["service_module"]

    result = None

    try:
        importlib.import_module(service_module)
    except ModuleNotFoundError as err:
        LOGGER.error("Failed to get module: %s", err)
        raise

    try:
        result = getattr(sys.modules[service_module], service)()
    except KeyError as err:
        LOGGER.error("Failed to get service class instance: %s", err)
        raise

    return result


def validate(config):
    """Test that provided json is a valid config."""
    if list(config.keys()) != [
        "port",
        "api_call_intervals",
        "log_level",
        "services",
    ]:
        raise ValueError(
            "Invalid config: config should contain port, api_call_intervals, log_level, services"
        )

    if not isinstance(config["port"], int):
        raise ValueError("Invalid config: port should be an integer")

    if not isinstance(config["api_call_intervals"], int):
        raise ValueError("Invalid config: api_call_intervals should be an integer")

    valid_log_levels = ["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG", "NOTSET"]
    if config["log_level"] not in valid_log_levels:
        raise ValueError(
            f"Invalid config log_lebel must be one of {', '.join(valid_log_levels)}"
        )

    if not isinstance(config["services"], dict):
        raise ValueError("Invalid config: services must be a dict of services")

    if len(config["services"]) == 0:
        raise ValueError("Invalid config: must contain at least one service")

    validate_services(config["services"])


def validate_services(services):
    """Test that each service in the config is a valid service configuration."""
    for service_name in services:
        validate_service(service_name, services[service_name])


def validate_service(service_name, service):
    """Test that a service contains a valid configuration."""
    expected_keys = [
        "service_class",
        "service_module",
        "api_url",
        "prometheus_metrics_mapping",
    ]

    if not all(key in service for key in expected_keys):
        raise ValueError(
            f"Invalid config: all services should contain '{', '.join(expected_keys)}'"
        )

    strings = ["service_class", "api_url", "secret_file_path"]

    for field in strings:
        if not isinstance(service[field], str):
            raise ValueError(f"Invalid config: {field} fields should be a string")

    validate_metrics_mapping(service_name, service["prometheus_metrics_mapping"])


def validate_metrics_mapping(service_name, metrics_mapping):
    """Test that the metrics mapping of a service is valid."""
    valid_metric_types = ["counter", "summary"]

    if not isinstance(metrics_mapping, dict):
        raise ValueError(
            f"Invalid config: {service_name} prometheus_metrics_mapping should be a dict"
        )

    if len(set(metrics_mapping)) != len(metrics_mapping):
        raise ValueError(
            f"Invalid config: {service_name} prometheus_metrics_mapping\
                        should contain only one of each metric type"
        )

    if not all(metric_type in valid_metric_types for metric_type in metrics_mapping):
        raise ValueError(
            f"Invalid config: {service_name} prometheus_metrics_mapping\
                        can only be one of {', '.join(valid_metric_types)}"
        )

    metrics = []

    for metric_type in metrics_mapping:
        metrics = metrics + metrics_mapping[metric_type]

    for metric in metrics:
        if not isinstance(metric, str):
            raise ValueError(
                f"Invalid config: {service_name} metric '{metric}' is not a string"
            )



"""Module to define what an API class should do and what it should look like."""
from datetime import datetime

from abc import ABC

import requests

from prometheus_client import Summary, Counter, Gauge, Histogram
from collectington.exceptions.collection_exceptions import UnsupportedPrometheusInstance


def enable_delta_metric(func):
    """
    This is a decorator that is used by metric methods. Since each method continuously
    adds total value to Prometheus counter, we do not want to double count metric value
    every time we get API data.

    This will keep track of previous metric and compare against new data to get delta value.
    Previous metric is defined as class attribute for each metric.

    This decorator simply returns the difference between new metric data and previous
    metric data.

    :return: delta of new metric data & previous data
    """

    def wrapper(self):
        # Assigning this to a new variable does not work as this variable will
        # reset value to 0. self is required as this will be called from a class

        previous_metric_name = f"previous_{func.__name__}"

        previous_metric_value = self.__class__._delta_metric_registry[
            previous_metric_name
        ]

        if previous_metric_value == func(self):
            return 0

        delta_metric = func(self) - previous_metric_value

        delta_metric_registry_copy = self.__class__._delta_metric_registry.copy()

        delta_metric_registry_copy[previous_metric_name] = func(self)

        setattr(self.__class__, "_delta_metric_registry", delta_metric_registry_copy)

        return delta_metric

    return wrapper


def register_metric_class(cls):
    """
    This is a class decorator to register metric function with a metric defined in config.
    This removes the process of defining  "get_metric" method like below:

        def get_metric(self, metric):
            if metric == "total_calls"
                return get_total_calls()
            elif metric == "total_missed"
                return get_total_missed()
            else:
                return

    Instead, service class methods that uses register_metric decorator will be auto-registered to
    metric_registry and will be mapped to the correct metric function.

    All subclasses of the AbstractApi class must use this decorator to successfully register all
    metric functions defined in its class.

    """
    cls._metric_registry = {}
    cls._delta_metric_registry = {}

    for methodname in dir(cls):
        method = getattr(cls, methodname)
        if hasattr(method, "_property"):
            cls._metric_registry.update({method._property[0]: methodname})
            cls._delta_metric_registry.update(
                {f"previous_{methodname}": 0}
            )  # this is for delta metric
    return cls


def register_metric(*args):
    """
    This is a class method decorator that registers a method to the registry.

    The argument is the name of a metric defined in config and this name will be mapped to
    the function using the decorator.
    """

    def wrapper(func):
        func._property = args

        return func

    return wrapper


class CollectingtonApi(ABC):
    """
    This class is an abstract class that includes implementations for common methods
    and forces its subclasses to implement abstract methods.

    The common methods include:
        - reading data via calling an api and cacheing the results
        - Invalidating cache after expiration period
        - Creating Prometheus instances based on config

    Any method can be overridden by its subclasses should a custom method is required.
    """

    def __init__(self):
        self.data_store = {}
        # Prometheus reads data 1 per minute(60 sec)
        self.data_store_expiration_sec = 60

        self.config = None
        self.headers = {}
        self.params = {}
        self.name_of_datastore = ""
        self.api_url = ""
        self.service_name = ""

        self.prometheus_metrics_mapping = {
            "counter": Counter,
            "gauge": Gauge,
            "histogram": Histogram,
            "summary": Summary,
        }

    def read_data(self, url, params, headers):
        """Request data from API.
        Update current read time.
        Return API response.
        """
        response = requests.request("GET", url, params=params, headers=headers)
        response = response.json()

        # this is required for invalidating cache after expiry
        self.data_store["data_read_at"] = datetime.now()

        return response

    def is_data_store_expired(self):
        """Check to see if data store has expired.
        Returns boolean.
        """
        data_cached_time = self.data_store["data_read_at"]

        time_delta = datetime.now() - data_cached_time

        return time_delta.total_seconds() >= self.data_store_expiration_sec

    def get_data_from_store(self, name_of_datastore):
        """
        Instead of having to call an API for every metric, different meteics can
        share the same response data(cache) before expiry.
        """
        if not self.data_store.get(name_of_datastore) or self.is_data_store_expired():
            response = self.read_data(self.api_url, self.params, self.headers)

            self.data_store[name_of_datastore] = response

        return self.data_store[name_of_datastore]

    def _init_p_method(self, p_method, api_metric):
        """Internal method to metric methods with labels only if they're provided."""
        if (
            self.config.get("prometheus_metric_labels") is not None
            and api_metric in self.config["prometheus_metric_labels"]
        ):
            labels = self.config["prometheus_metric_labels"][api_metric]
            return p_method(api_metric, api_metric, labels)

        return p_method(api_metric, api_metric)

    def generate_prometheus_metric_instances(self):
        """Create a list of metrics for service."""
        list_of_metric_instances = []

        for p_metric, api_metrics in self.config["services"][self.service_name][
            "prometheus_metrics_mapping"
        ].items():
            p_method = self.prometheus_metrics_mapping[p_metric]

            p_instances = [
                self._init_p_method(p_method, api_metric) for api_metric in api_metrics
            ]

            list_of_metric_instances += p_instances

        return list_of_metric_instances

    def call_prometheus_metrics(self, service_metric_dict, list_of_metric_instances):
        """Handles sending the metric data to prometheus.

        For metrics that have labeled data, we loop through all the labels and
        send each one sequentially.

        For metrics with no label data, we make a single upload for the value.
        """
        for p_instance in list_of_metric_instances:
            metric = str(p_instance).split(":")[1]

            if (
                self.config.get("prometheus_metric_labels") is not None
                and metric in self.config["prometheus_metric_labels"]
            ):
                for labels_and_metric_object in service_metric_dict[metric]:
                    label_list = self.config["prometheus_metric_labels"][metric]
                    labels, val = self._split_labeled_metric_dict(
                        labels_and_metric_object, label_list
                    )
                    self._update_metric(p_instance.labels(*labels), val)
            else:
                self._update_metric(p_instance, service_metric_dict[metric])

    @staticmethod
    def _split_labeled_metric_dict(labeled_metric_dict, label_list):
        """Split single dict into list of labels and the metric value."""
        labeled_metric = labeled_metric_dict.copy()

        labels = []
        for key in label_list:
            labels.append(labeled_metric[key])
            labeled_metric.pop(key, None)

        val = list(labeled_metric.values())[0]

        return labels, val

    @staticmethod
    def _update_metric(p_instance, val):
        """
        Prometheus client includes 4 metrics: Counter, Gauge, Histogram and Summary.
        Each metric needs to be instantiated and each metric calls a method that generates
        outputs in a format that Prometheus can read.

        Example methods include: inc(), dec(), set(), observe() and etc - check documentation

        User of this class can override this method to determine which Prometheus method
        will be used for each metric
        """
        if isinstance(p_instance, Counter):
            # inc is a method from Prometheus client
            p_instance.inc(val)
        elif isinstance(p_instance, (Summary, Histogram)):
            p_instance.observe(val)
        elif isinstance(p_instance, Gauge):
            p_instance.set(val)
        else:
            raise UnsupportedPrometheusInstance

    def get_metric(self, metric):
        """
        This method takes an argument which is a metric name defined in config and will
        key in the value to metric_registry dictionary to get the correct metric function.
        """

        try:
            metric = self.__class__._metric_registry[metric]
            metric_func = getattr(self.__class__, metric)
            return metric_func(self)
        except (IndexError, KeyError):
            # Certain errors occur due to issues with API calls.
            return None



"""Test that the config parsing module is operating as expected."""
import unittest

from json.decoder import JSONDecodeError
from collectington.config import parse, validate


class TestConfigParse(unittest.TestCase):
    """Test that the config parsing function is operating as expected."""

    def test_parse_valid_json(self):
        """Test that function parses json correctly."""
        test_json = '{"json": "test", "one": 1}'
        expected = {"json": "test", "one": 1}

        output = parse(test_json)

        self.assertEqual(output, expected)

    def test_parse_fails_with_invalid_json(self):
        """Ensure the correct Error is thrown if input isn't valid json."""
        test_string = "invalid json"

        with self.assertRaises(JSONDecodeError):
            parse(test_string)

    def test_parse_fails_with_non_string_input(self):
        """Ensure that a TypeError is thrown if a string isn't provided as an argument."""
        test_input = 3

        with self.assertRaises(TypeError):
            parse(test_input)


class TestConfigValidateConfig(unittest.TestCase):
    """Test that configs are validated correctly."""

    def test_parse_valid_config(self):
        """Test that a valid config will not raise any exceptions."""
        with open(
            "collectington/test/files/config/config.json", encoding="utf-8"
        ) as file:
            valid_config = file.read()

        parsed_config = parse(valid_config)

        try:
            validate(parsed_config)
        except ValueError as err:
            self.fail(f"Exception raised:\n{err}")


if __name__ == "__main__":
    unittest.main()



import setuptools
from distutils.core import setup


setup(
    name="collectington",
    packages=setuptools.find_packages(),
    install_requires=["prometheus-client", "termcolor", "pyfiglet", "requests"],
    scripts=["cton"],
    license="MIT",
    version="0.1.0",
    description="Collectington is a framework that allows any 3rd party API data to be sent to Prometheus",
    author="YoungJun Park, Nathan Boswell, Eric Jalbert",
    author_email="yjpark@homex.com, nboswell@homex.com, ejalbert@homex.com",
    url="https://github.com/HomeXLabs/collectington",
    download_url="https://github.com/HomeXLabs/collectington",
    package_dir={"collectington": "collectington"},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
    ],
    python_requires=">=3.6",
)



"""Module for printing ascii introduction."""
from termcolor import colored
from pyfiglet import figlet_format


def print_ascii():
    """Function to print ascii introduction."""
    result = colored(figlet_format("Welcome! \n This is Collectington"), color="green")

    print(result)



"""Module that contains function to create a logger object."""
import logging

LOG_MSG_FORMAT = "%(asctime)s|%(name)-12s|%(levelname)-8s|%(message)s"
LOG_DATE_FORMAT = "%Y/%m/%d|%H:%M:%S"
LOGGING_LEVEL = "INFO"
DATA_COLLECTION_LOGGER = "DATA-COLLECTION"


def setup_logging():
    """Return logger object."""
    log_formatter = logging.Formatter(LOG_MSG_FORMAT, LOG_DATE_FORMAT)

    logger = logging.getLogger(DATA_COLLECTION_LOGGER)
    logger.setLevel(LOGGING_LEVEL)

    console_handler = logging.StreamHandler()
    console_handler.setFormatter(log_formatter)

    logger.addHandler(console_handler)
    logger.handlers[:] = [console_handler]

    return logger
